// Egoboo - Ticker.c

#include "Ticker.h"
#include "Clock.h"
#include <assert.h>
#include <stddef.h>

void ticker_initWithInterval(Ticker *ticker, double interval)
{
	assert(ticker != NULL && "ticker_initWithInterval: NULL ticker passed!");
	if(ticker == NULL || interval <= 0) return;
	
	ticker->lastTime = clock_getTime();
	ticker->numTicks = 0;
	ticker->tickInterval = interval;
}

void ticker_initWithFrequency(Ticker *ticker, int freq)
{
	double interval = 1.0 / freq;
	ticker_initWithInterval(ticker, interval);
}

void ticker_update(Ticker *ticker)
{
	double deltaTime, currentTime;
	assert(ticker != NULL && "ticker_update: NULL ticker passed!");
	if(ticker == NULL) return;

	currentTime = clock_getTime();
	deltaTime = currentTime - ticker->lastTime;

	while(deltaTime > ticker->tickInterval)
	{
		ticker->numTicks++;
		ticker->lastTime += ticker->tickInterval;

		deltaTime -= ticker->tickInterval;
	}
}

int ticker_tick(Ticker *ticker)
{
	int numTicks;

	assert(ticker != NULL && "ticker_tick: NULL ticker passed!");
	if(ticker == NULL) return 0;

	numTicks = ticker->numTicks;
	ticker->numTicks--;
	return numTicks;
}
